#!/usr/bin/env -S runhaskell --ghc-arg="-i lib"
{-# LANGUAGE NamedFieldPuns, OverloadedStrings #-}

--import Control.Map
import Data.Attoparsec.Text (parseOnly)
import Data.List as List (foldl')
import Data.Map as Map (Map, elems, empty, insertWith, toList)
import Data.Metadata (Article(..), Tome, Tomes, byTome)
import Data.Set as Set (Set, singleton, union)
import Data.Text as Text (Text, concat, intercalate, splitOn, unpack)
import Data.Text.IO as Text (readFile)
import System.Directory (listDirectory)
import System.Environment (getArgs, getProgName)
import System.Exit (die)
import System.FilePath ((</>), (<.>), dropTrailingPathSeparator, takeBaseName)
import System.Script (syntax)
import Text.XML.HXT.Core (
    (>>>), ($<), Arrow, ArrowIO, ArrowXml, IOStateArrow, XmlTree, arr, arrIO0, attr, constL, getChildren, mkelem, readDocument, runX, selem, txt, unlistA, withValidate, writeDocument
  )

{-
getTome :: FilePath -> String
getTome = drop 1 . takeBaseName . dropTrailingPathSeparator
-}

type IOArrow b x = IOStateArrow () b x

compileTome :: FilePath -> FilePath -> (Int, Tome) -> IOArrow b XmlTree
compileTome source target (tomeNumber, metadata) =
  tomeXml >>> writeDocument [] (target </> "T" <> strTome <.> "xml")
  where
    strTome = show tomeNumber
    tomeXml = selem "/"
                [selem "teiCorpus"
                  [teiHeader
                    (titleStmt strTome)
                    (corpusPublicationStmt)
                    (corpusSourceDesc strTome)
                  ,(articleFrom source $< constL (Map.elems metadata))]]

        --, constL (articleFrom source <$> metadata) >>> undefined]]

articleFrom :: FilePath -> Article -> IOArrow b XmlTree
articleFrom source (Article {tome, uid, headWord, author, artflDomains, enccreDomains}) =
  selem "TEI"
    [teiHeader
      (selem "titleStmt" [title $ Text.unpack headWord])
      (publicationStmt
        [selem "p" [txt "Annotated with Stanza by project GEODE"]])
      (sourceDesc [selem "author" [txt $ Text.unpack author]])
    , mkelem "text" metadata [body]]
  where
    path = source </> "T" <> show tome </> Text.unpack uid <.> "tei"
    body = readDocument [withValidate False] path >>> getChildren
    key .= value = attr key (txt $ Text.unpack value)
    aDomainsAttr = Text.concat [":", Text.intercalate ":" artflDomains, ":"]
    eDomainsAttr = Text.concat [":", Text.intercalate ":" enccreDomains, ":"]
    metadata =
      ["uid" .= uid, "author" .= author, "artflDomains" .= aDomainsAttr, "enccreDomains" .= eDomainsAttr, "book" .= "LGE"]

--    arrIO0 (listDirectory source)
--  >>> unlistA
--  >>> arr (source </>)
--  >>> readFromDocument [withValidate no]

{-
getTEICorpus :: ArrowXml a => String -> a b XmlTree -> a b XmlTree
getTEICorpus tome corpus = selem "/" [selem "teiCorpus" [teiHeader tome, corpus]]
-}

teiHeader :: ArrowXml a =>
  a b XmlTree -> a b XmlTree -> a b XmlTree -> a b XmlTree
teiHeader aTitle aPublication aSource =
  selem "teiHeader"
    [selem "fileDesc" [aTitle, aPublication, aSource]]
    --selem "fileDesc" [titleStmt tome, publicationStmt, sourceDesc tome]
  --]

titleStmt :: ArrowXml a => String -> a b XmlTree
titleStmt tome = selem "titleStmt" [
      title tome
    , selem "respStmt" [
            selem "resp" [txt "Digitized by"]
          , selem "orgName" [txt "Bibliothèque Nationale de France"]
          , selem "resp" [txt "Encoded by"]
          , selem "orgName" [txt "ICAR"]
        ]
  ]

publicationStmt :: ArrowXml a => [a b XmlTree] -> a b XmlTree
publicationStmt content = selem "publicationStmt" content

corpusPublicationStmt :: ArrowXml a => a b XmlTree
corpusPublicationStmt =
  (publicationStmt
    [selem "distributor"
      [selem "orgName" [txt "Project GÉODE"]
      ,selem "address"
          [selem "addrline" [txt "ICAR UMR 5191"]
          ,selem "addrline" [txt "ENS de Lyon"]]]])

sourceDesc :: ArrowXml a => [a b XmlTree] -> a b XmlTree
sourceDesc biblContent =
  selem "sourceDesc" [selem "bibl" biblContent]

corpusSourceDesc :: ArrowXml a => String -> a b XmlTree
corpusSourceDesc tome =
  sourceDesc 
    [title tome
    ,selem "author" [txt "Collective"]
    ,selem "creation"
      [selem "date" [txt "1885"]]
    ,selem "imprint"
      [selem "date" [txt "1885"]
      ,selem "publisher" [txt "H. Lamirault et Cie,"]
      ,selem "pubplace" [txt "Paris"]]
    ,mkelem "biblScope" [attr "unit" (txt "volume")] [txt tome]]

title :: ArrowXml a => String -> a b XmlTree
title titleContent = selem "title" [txt titleContent]

{-
$ unlines [
      "La Grande Encyclopédie, Inventaire raisonné des sciences, des lettres et"
    , "des arts par une société de savants et de gens de lettres. Tome " ++ tome
  ]]
-}

{-
getTomes :: ArrowIO a => FilePath -> a b Tome
getTomes metadata = constA undefined

getTomes :: FilePath -> IO Tomes
getTomes =
  fmap (foldl' index Map.empty . fmap fields . drop 1 . lines) . Text.readFile
  where
    fields = Text.splitOn ","
    index tomes [file, tome, rank, head, domain] =
      Map.insertWith Set.union tome (Set.singleton ) tomes
    index tomes _ = tomes
    
-}

main :: IO ()
main = getArgs >>= cli
  where
    cli [sourceMeta, sourceRoot, targetRoot] =
        Text.readFile sourceMeta
      >>= pure . parseOnly byTome
      >>= either
            die
            (mapM_ (runX . compileTome sourceRoot targetRoot) . Map.toList)
    cli _ = syntax "METADATA_CSV_FILE SOURCE_DIR TARGET_DIR"
